// MINIROFF Version 2.0.

//      Copyright 1992, 1993 by Greg Fife
//
//      All rights are reserved.  This program may not be distributed
//      without permission of the Computer Science Department of the
//      University of Virginia.
//
// This software is provided on an "as-is" basis.  NO guarantee is made
// that this software is suitable for any purpose.

// Standard C++ headers used by Miniroff

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>


// Miniroff Constants:

const int  True     = 1 ;
const int  False    = 0 ;

const char Newline  = '\n' ;
const char Period   = '.'  ;
const char Plus     = '+'  ;
const char Minus    = '-'  ;
const char Space    = ' '  ;

const int Max_Line_Width = 256 ;       // Limit for input and output
                                       // line buffers

const int Argument_Position = 3 ;     // an argument value on a "directive
                                      // line" starts on the 3rd character

// The page geometry defaults assume a 10 character/inch font

const int Default_Left_Margin  = 10 ;  // 1 inch left margin
const int Default_Right_Margin = 10 ;  // 1 inch right margin
const int Default_Line_Width   = 65 ;  // 65 char print zone for 8.5" wide
                                       // paper

const int Default_Para_Indent  =  5 ;  // 0.5 inch indent for a paragraph


// Miniroff Types

typedef char charbuffer[Max_Line_Width+1] ;
                                      // A "charbuffer" is an array of
                                      // characters.  The +1 is to leave
                                      // space for a trailing '\0'  


// Miniroff Global Variables:

int left_margin = Default_Left_Margin  ;   // Left margin in characters
int line_width  = Default_Line_Width   ;   // Output line length.
int right_margin= Default_Right_Margin ;   // Right margin in characters

int page_width  = Default_Left_Margin + Default_Line_Width +
                  Default_Right_Margin ;   // Total width of page

int para_indent = Default_Para_Indent  ;   // Indent 1st line of paragraph

charbuffer output_buffer ;              // Buffer for one output line
int  chars_buffered = 0 ;               // Number of characters in buffer
int  word_pos = 0 ;                     // Index of the space most recently
                                        //   saved in the buffer - marks end
                                        //   of last complete word
int line_limit = Default_Line_Width ;   // Number of characters to accept
                                        //   before we flush.

// Document status Flags

int first_line  = False ;        // Is this the first line of a paragraph?


// Miniroff Functions and Procedures :

// The following are "function prototypes."  They indicate the return
// type and parameters of a function.  This allows a compiler to handle
// a call to a function that occurs before the function is actually
// defined.

// See "C - How To Program" by Deitel & Deitel, page 142

int readline () ;                // Read 1 line of the input stream
                                 // Return True if the end of the
                                 // input was not seen

void readtext (charbuffer input_line) ;
                                 // Read 1 line of normal input text
                                 // input_line - text to save

void buffer_char (const char store) ;  // Store one character in output buffer
                                       // store - character to save

void flush_buffer () ;           // Write buffered characters to
                                 // one line of the output - start a new
                                 // line

void start_paragraph () ;        // Write buffered characters and start
                                 // a new paragraph


void set_parms () ;              // Adjust buffer values after
                                 // margins have changed

void format_line (charbuffer line, const int nchars, const int indent,
                  const int width, const int fill) ;
                                 // format and output a line of text.
                                 // line   - the buffer of characters to print
                                 // nchars - number of characters in line[]
                                 // indent - initial indentation of the line
                                 // width  - number of characters in line
                                 //          after indentation
                                 // fill   - True if line should be "filled"
                                 //          to the entire width


void new_line () ;              // send a '\n' to the output stream

void directive (charbuffer input_line) ;
                                // Read and interpret an input line
                                // containing a miniroff directive.
                                // input_line - buffer of input characters

int dir_adjust (charbuffer input_line, const int value) ;
                                 // calculate a Miniroff page value
                                 // input_line - line containing the
                                 //              directive
                                 // value - old value of the page value
                                 // returns the new value for the page value

void adjust_globals (charbuffer input_line) ;
				 // reset page parameters after a global
				 // (left margin, right margin, etc.) change
				 //
				 // input_line - line containing the current
				 //              MINIROFF directive

void dir_br (charbuffer input_line) ;
                                 // handle a break (.br) directive
                                 // input_line - buffer of input characters

void dir_pp (charbuffer input_line) ;
                                // handle a new paragraph (.pp) directive
                                // input_line - buffer of input characters

void dir_sp (charbuffer input_line) ;
                                // handle a blank space (.sp) directive
                                // input_line - buffer of input characters


int read_integer (charbuffer input_line, const int start, char &leading) ;
                               // interpret the integer following an
                               // adjust value directive.  Verify
                               // that nothing follows the integer
                               // except a newline.
                               // input_line - buffer of input characters
                               // start - begin examing buffer at this
                               //         position
                               // leading - return a '+' if the integer
                               //           was preceeded by a '+', return
                               //           a '-' if the integer was
                               //           preceeded by a '-', return
                               //           ' ' otherwise.
                               // returns: the integer from the input_line

void check_end (charbuffer input_line, const int start = Argument_Position ) ;
                               // verify that nothing follows
                               // except whitespace
                               // input_line - buffer of input characters
                               // start     - start examining at this
                               //             position

void abend( char *error, char *line = 0 ) ;
                              // print a 1-parameter error message and exit
                              // error    - the specific error type to be
                              //            displayed
                              // line     - the input line to be displayed
                              //            with the message.  Default is
                              //            "nothing" - do not print a line

////////////////////////////////////////////////////////////////
//
//     main
//
//     Call the readline() routine until the end of input is seen.
//     Flush output line buffer after all input has been read.
//

main ()
{
    while ( readline () ) {   // empty loop body - all the real work
    }                         //   is done by readline
    flush_buffer () ;         // print any remaining characters - last line
                              // of output
    return 0 ;
}


/////////////////////////////////////////////////////////////////////
//
//     readline
//
//     read 1 line of the input stream, gather it into an input buffer.
//     pass this buffer to readtext() for normal input text, or pass it
//     to directive() if it is a miniroff directive
//

int readline ()
{
    charbuffer input_buffer ;    // this array of characters contains 1
                                 // input line.  A \0 (a "null") at a
                                 // character position will mark the end
                                 // of the line

    int chars_seen = 0 ;                 // Number of characters read
    int eof_flag = False ;               // Have we seen the end of input?
    char char_read ;                     // One character from input stream

    while (True) {                       //  loop is ended by one of two
                                         //  "break" statements below
        if (!cin.get(char_read)) {                  //  End Of File
            input_buffer[chars_seen] = 0 ;
            eof_flag = True ;
            break ;
        }

        if (char_read==Newline) {                   // End of Line Mark
            input_buffer[chars_seen] = 0 ;
            break ;
        }

  // Otherwise - a character has been read

          if (chars_seen==Max_Line_Width) {
	      abend ("Input Line to large") ;
	  }
          input_buffer[chars_seen++] = char_read ;
                                    // Store character and increment index
    }  // while loop

    if (input_buffer[0]==Period) {  // If the first character is a period,
                                    // this is a miniroff directive
        directive(input_buffer) ;
    }
    else {
        readtext(input_buffer) ;
    }

    return !eof_flag ;  // Return true if eof not seen
}


//////////////////////////////////////////////////////////////////////////
//
//  readtext
//
//  Read a line of normal input, and pass it into bufferchar() to be saved
//  until we are ready to write an output line.
//
//  We only want to store 1 blank space between words.  If multiple blanks
//  or a combination of blanks, tabs, and newlines separates two words,
//  then only a single space should be stored with bufferchar()
//
//  input_line - text read in by caller. This is a "null-terminated string";
//               a \0 value in a character position marks the end of the line
//

void readtext (charbuffer input_line)
{
    int within_word = False ;     // Are we in the middle of a word?
    int char_index = 0 ;          // index the input_line array
    char char_read ;              // the current character of the array

// The loop below stops because "input_line" is null-terminated.
// In other words, the end of the line is marked with a 0

// The assignment "a=b" is also an expression - it can be used inside an
// if () or a while().  The VALUE of the expression is the VALUE that is 
// assigned into the variable to the left of the '=' mark.  Thus,
// (x=2) is an expression with a VALUE 2, and it has the SIDE EFFECT of
// placing the value 2 in the variable x.

    while (char_read=input_line[char_index++]) {
        if (isspace(char_read)) {     // whitespace - separates words
            if (within_word) {
                within_word = False ;  // No longer within a word
                buffer_char (Space) ;  // buffer a single blank space
            }
            else {
            }                // Ignore - a space has already been saved
        }
        else  {    // not whitespace - part of word
            buffer_char(char_read) ;
            within_word = True ;
        }
    }  // while loop

// The '\n' at the end of the input line is a whitespace.  Handle it.

    if (within_word) {
        buffer_char (Space) ;
    }
    return ;
}


///////////////////////////////////////////////////////////////////
//
//     buffer_char
//
//     Store a single character in the output buffer.  Flush the
//     output buffer if we reach the current line limit.
//
//     store - character to save in output buffer
//

void buffer_char (const char store)
{
    output_buffer[chars_buffered] = store ;
    if (store == Space) {
        word_pos = chars_buffered ;  // keep track of last complete word
    }
    chars_buffered ++ ;
    if (chars_buffered == line_limit) {
        flush_buffer () ;           // write out 1 line
    }
}


//////////////////////////////////////////////////////////////////////
//
//     flush_buffer
//
//     send the line buffer to format_line to be formatted and printed
//     keep any partial last word in the beginning of the buffer for
//     the next output line.
//

void flush_buffer ()
{
    int char_index ;      // index for character array
    int chars_left ;      // number of characters left in buffer 
                          // past end of last word
    int indent ;          // left indentation for this particular line

    if (first_line == True) {
        indent = left_margin + para_indent ;
    }
    else {
        indent = left_margin ;
    }

// Format Line, Save any unused characters, and set new chars_buffered

    if (chars_buffered < line_limit)  {   // We haven't filled up a line
        format_line(output_buffer,chars_buffered,indent,line_limit,False);
        chars_buffered = 0 ;
    }
    else {
        if (word_pos==0) {  // We don't have a complete word
            abend ("Word length exceedes line length");
        }
        format_line(output_buffer,word_pos,indent,line_limit,True);
        if (word_pos==chars_buffered - 1)   {    // Last Word was complete
            chars_buffered=0 ;
        }
        else {
            chars_left = chars_buffered - word_pos - 1 ; 
                    // Do not count space after last complete word
            for (char_index = 0 ; char_index < chars_left ; char_index++) {
                output_buffer[char_index] = 
                  output_buffer[char_index+word_pos+1] ;
            }
            chars_buffered = chars_left ;
        }
    }

// reset remaining buffer variables for the next line.
    line_limit = line_width ;
    word_pos = 0 ;
    first_line = False ;   // Not the Momma!
}


/////////////////////////////////////////////////////////////////////
//
//     start_paragraph
//
//     Begin a new paragraph.  Flush existing buffered output, set
//     the flag, and adjust the line limit.
//

void start_paragraph ()
{
    first_line = True ;
    line_limit = line_limit - para_indent ; 
                               // We have less space for this line.
    assert(line_limit>0) ;
}


////////////////////////////////////////////////////////////////
//
//     set_parms
//
//     Adjust buffer values after global variables have been
//     changed.  Check for buffer overflow.
//

void set_parms ()
{
    assert (chars_buffered == 0) ;   // should already have flushed buffer

    if (first_line) {
        line_limit = line_width - para_indent ;
    }
    else {
        line_limit = line_width ;
    }

    if (line_limit >= Max_Line_Width) {
        abend ("Print zone exceedes output buffer size.") ;
    }
}

//////////////////////////////////////////////////////////////////////
//
//     directive
//
//     Interpret a miniroff directive line, and branch to the appropriate
//     routine to handle the  command.
//
//     input_line - buffer of characters containing the miniroff directive
//

void directive (charbuffer input_line)
{
    char first_char  ;    // first character of the miniroff directive
    char second_char ;    // second character of the miniroff directive

    first_char=input_line[1] ;
    second_char=input_line[2] ;

    if (first_char==0 || second_char==0) {
        abend ("Syntax Error.",input_line);
    }

// Branch to appropriate routine to handle directive

    if (first_char=='l' && second_char=='m') {         // .lm
        left_margin = dir_adjust(input_line,left_margin) ;
	adjust_globals (input_line) ;
    }
    else if (first_char=='r' && second_char=='m') {    // .rm
        right_margin = dir_adjust(input_line,right_margin) ;
	adjust_globals (input_line) ;
    }
    else if (first_char=='p' && second_char=='w') {    // .pw 
        page_width = dir_adjust (input_line,page_width) ;
	adjust_globals (input_line) ;
    }
    else if (first_char=='p' && second_char=='i') {    //  .pi
        para_indent = dir_adjust(input_line,para_indent) ;
	adjust_globals (input_line) ;
    }
    else if (first_char=='b' && second_char=='r') {    //  .br
        dir_br(input_line) ;
    }
    else if (first_char=='s' && second_char=='p') {    //  .sp
        dir_sp(input_line) ;
    }
    else if (first_char=='p' && second_char=='p') {    //  .pp
        dir_pp(input_line) ;
    }
    else {
        abend ("Syntax Error.",input_line) ;
   }
}


////////////////////////////////////////////////////////////
//
//     dir_adjust
//
//     Calculate the new value for a Miniroff variable
//     from the old value and an "adjust" directive line.
//
//      input_line - array containing the directive
//      old_value  - previous value of the quantity being changed
//

int dir_adjust (charbuffer input_line, const int old_value)
{
    int delta ;                 // amount for value to change
    char prefix ;               // +,-, or ' ' before the change
    int new_value = old_value ; // new value of the quantity

    delta=read_integer(input_line, Argument_Position, prefix) ;

    flush_buffer () ;

    if (prefix==Plus) {
        new_value += delta ;
    }
    else if (prefix==Minus) {
        new_value -= delta ;
    }
    else {
        new_value = delta ;
    }


    if (new_value <= 0) {
        abend ("Underflow.",input_line) ;
        exit(1) ;
    }

    return new_value ;
}


/////////////////////////////////////////////////////////////////
//
//     adjust_globals
//
//     adjust and test global values after a dir_adjust call
//
//     input_line - line containing the "adjust value" command
//
void adjust_globals(charbuffer input_line)
{
//  calculate a new line width, given that a margin has just changed

    line_width = page_width - left_margin - right_margin ;

// make sure we haven't made anything important turn into a negative number

    if (line_width <= 0) {
	abend ("Underflow.",input_line) ;
    }

    if (para_indent >= line_width) {
        abend ("Indent exceedes print zone length",input_line) ;
    }

    set_parms () ;  // update the output buffer values
    return ;
}


/////////////////////////////////////////////////////////////////
//
//     dir_br
//
//     Execute a "break line" (.br) directive
//
//     input_line - buffer of characters containing the directive
//

void dir_br (charbuffer input_line)
{
    check_end (input_line) ;
    flush_buffer () ;
}


//////////////////////////////////////////////////////////////////
//
//     dir_pp
//
//     Execute a "new paragraph" (.pp) directive
//
//     input_line - buffer of characters containing the directive
//

void dir_pp (charbuffer input_line)
{
    check_end (input_line) ;
    flush_buffer () ;
    start_paragraph () ;
}


/////////////////////////////////////////////////////////////////////
//
//     dir_sp
//
//     Execute a "blank space" (.sp) directive
//
//     input_line - buffer of characters containing the directive
//

void dir_sp (charbuffer input_line)
{
    check_end (input_line) ;
    flush_buffer () ;
    new_line () ;
}


//////////////////////////////////////////////////////////////////
//
//     read_integer
//
//     Read the leading + or - and the integer for an "adjust value"
//     directive
//
//     input_line - buffer of characters containing the directive
//                  This is a "null terminated" string - a true 0
//                  (\0) in a character position marks the end of a line
//     start      - begin searching at this position
//     leading    - return the '+' or '-' if one is found.  Otherwise,
//                  return a ' '
//

int read_integer (charbuffer input_line, const int start, char &leading)
{
    int char_index = start ;  // position in array
    int value = 0 ;           // value of the integer

    leading = Space ;

// skip leading spaces
    while (input_line[char_index]==Space) {
        char_index++ ;
    }

// handle any leading + or -

    if (input_line[char_index]==Plus || input_line[char_index]==Minus) {
        leading=input_line[char_index++] ;
        while (input_line[char_index]==Space) {
            char_index++ ;   // skip spaces after + or -
            }
    }

    if (!isdigit(input_line[char_index])) {
        abend ("Syntax Error.", input_line) ;
    }

    while (isdigit(input_line[char_index])) {
        value *= 10 ;
        value += input_line[char_index] - '0' ;    // '0' - '0' is 0,
                                                   // '1' - '0' is 1, 
                                                   // and so forth
        char_index ++ ;
    }

    check_end (input_line,char_index) ;  // Make sure nothing follows
    return value ;
}


////////////////////////////////////////////////////////////////
//
//     check_end
//
//     Check the characters after the directive, +/- (if any), and
//     integer (if any).  There should be blank spaces only or nothing
//     until the end of the line
//
//     input_line - buffer of characters containing the directive
//                  This is a "null-terminated" string - a true 0
//                  (\0) in a character position marks the end of the line
//     start      - begin scan here
//
//     NOTE: in the function prototype above, the "start" parameter is
//           given a default 
//           value of "Argument Position" - the first character after the
//           two letters of a miniroff directive.
//

void check_end (charbuffer input_line, const int start)
{
    int char_index=start ;  // position within array
    char onechar ;          // one character of the array

// The assignment inside the while() loop condition below is also an
// expression.  See the comment in read_text() for an explanation

// loop below stops because the end of the line in input_line is a 0

    while (onechar=input_line[char_index++]) {
        if (!isspace(onechar)) {
            abend ("Extra characters after directive.",input_line) ;
        }
    }
}


///////////////////////////////////////////////////////////////
//
//     format_line
//
//     Send an output buffer to the output stream.  Insert spaces
//     between words as needed to "fill out" to the right margin
//
//     line    - the array of characters to print
//     nchars  - number of characters in line[]
//     indent  - initial indentation of the line
//     width   - number of characters in line after indentation
//     fill   -  True if line should be "filled" to the entire width

//  NOTE - because of the high level of nesting in this function, an
//  indentation of 2 rather than 4 is used

void format_line (charbuffer line, const int nchars,
		  const int indent, const int width, const int fill)
{
  int real_nchars ;     // number of characters in buffer, ignoring trailing
                        // spaces
  int space_index ;     // count out spaces for left indent
  int char_index ;      // index characters through array
  int gap_total ;       // number of gaps (spaces) between words
  int gap_count ;       // count the gaps as they are printed out
  int total_spaces ;    // total spaces needed to pad line
  int spaces_per_gap ;  // number of spaces in each gap
  int extra_spaces ;    // number of extra spaces to finish line
  int extra_space_sep ; // separation between gaps that require extra spaces

  real_nchars = nchars ;

// ignore any trailing spaces in the output buffer

  while (real_nchars > 0 && line[real_nchars-1] == Space) {
    real_nchars-- ;
  }

  if (real_nchars==0) {
    return ;  // nothing to format
  }

// count the gaps
  gap_total = 0 ;
  for (char_index = 0 ; char_index < real_nchars ; char_index++) {
    if (line[char_index]==Space) {
      gap_total++ ;
    }
  }

// print the left margin

  for (space_index=0 ; space_index<indent ; space_index++) {
    cout << Space ;
  }

  if (!fill || !gap_total)  {    // no fill required, or no gaps to fill
    for (char_index = 0 ; char_index < real_nchars ; char_index++) {
       cout << line[char_index] ;
    }
  }
  else {
    total_spaces = width - (real_nchars-gap_total) ;
    spaces_per_gap = total_spaces / gap_total ;
    extra_spaces   = total_spaces % gap_total ;

// if extra_spaces is 0, do not calculate a separation 
// (do not divide by 0)

    if (extra_spaces) {
      extra_space_sep = gap_total / extra_spaces ;
    }

    gap_count = 0 ;
    for (char_index = 0 ; char_index < real_nchars ; char_index++) {
      cout << line[char_index] ;
      if (line[char_index]==Space) {
        gap_count ++ ;
           
// print the spaces found in ALL gaps.  loop starts at 1
// because first space has already been printed
 
        for(space_index=1 ;space_index<spaces_per_gap; space_index++) {
            cout << Space ;
          }

 // if one extra space is needed for this particular gap, print it

        if (extra_spaces && ! (gap_count % extra_space_sep)) {
          cout << Space ;
          extra_spaces -- ;   // stop when we have printed all the extra
        }
      }    // if (line[char_index]==Space)
    }      // for (char_index = 0   
  }        // else
  new_line() ;
}


/////////////////////////////////////////////////////////////////
//
//     new_line
//
//     Send a newline to the output stream
//

void new_line ()
{
    cout << Newline ;
}


////////////////////////////////////////////////////////////////////
//
//     abend
//
//     Abnormal Program Ending - print a useful message and exit.
//
//     message   -  the error message to be printed
//     line      -  the line upon which the error was found, if any.
//                  This parameter is given a default value of 0 (no line)
//                  in the function prototypes above.
//

void abend ( char *message , char *line)
{
    cerr << "Abnormal Ending.  " << message << "\n" ;
    if (line != 0) {
        cerr << "***  " << line << "\n" ;
    }
    exit (1) ;
}
